// -*- mode: c++ -*-
// $Id: dualmesh.h,v 1.10 2000/08/08 21:28:16 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/dualmesh.h,v $

#ifndef __DUALMESH_H__
#define __DUALMESH_H__

#include "cvec4t.h"
#include "subtri.h"
#include "tagmesh.h"
#include "grid3d.h"

class OcTree;
class Silhouettes;

class BoundaryEdge {
public:
  BoundaryEdge(Tri *t, EnoType e) :
    _tri(t), _eno(e), _normal(t->normal()) { };
  ~BoundaryEdge() { _tri = 0; };

  Tri* tri() const { return _tri; };
  EnoType eno() const { return _eno; };
  Vertex* v1() const { return _tri->headVert(_eno); };
  Vertex* v2() const { return _tri->tailVert(_eno); };
  
private:
  Tri *_tri;
  EnoType _eno;
  cvec3d _normal;
};

// projection of mesh's internal edges in the dual space
class DualEdge {
public:
  // pos1 and pos2 should be projected to the same surface of the 4D cube
  DualEdge(Tri *t1, Tri *t2, EnoType e1, EnoType e2, bool fac,
	   int tp, const cvec4d& pos1, const cvec4d& pos2);
  ~DualEdge() { };

  // the two triangles that share the edge
  Tri* tri(int i) const { return _tri[i]; };
  // the edge number in the two triangles
  EnoType eno(int i) const { return _eno[i]; };
  // end point positions in the dual space
  const cvec3d& v(int i) const { return _dv[i]; };
  // front facing or back facing
  bool facing() const { return _facing; };

  // silhouette checked tag
  bool checked() const { return _checked; };
  void check()   { _checked = true; _checkedEdges.push_back(this); };
  void uncheck() { _checked = false; };

  // touch a box
  bool touch(const cvec3d& minp, const cvec3d& maxp) const;

  // project edges to dual space
  static void project(OcTree *dmesh, Tri *tri, EnoType eno);

  // clear silhouette checked tags
  static void clearCheckTags() {
    int csz = _checkedEdges.size();
    for (int i = 0; i < csz; i++ )
      _checkedEdges[i]->uncheck();
    _checkedEdges.clear();
  };

private:
  Tri *_tri[2];
  EnoType _eno[2];
  cvec3d _dv[2];
  bool _facing;

  bool _checked;
  static vector<DualEdge*> _checkedEdges;
};

// octree node
class OcTree {
public:
  OcTree() : _depth(0), _min(-1, -1, -1), _max(1, 1, 1), _isLeaf(true)
    { for (int i=0; i<8; i++) _child[i] = 0; };
  OcTree(const cvec3d& minp, const cvec3d& maxp, int depth = 0) :
    _depth(depth), _min(minp), _max(maxp), _isLeaf(true)
    { for (int i=0; i<8; i++) _child[i] = 0; };
  ~OcTree();

  // depth in octree
  int depth() const { return _depth; };

  // bounding box
  const cvec3d& minp() const { return _min; };
  const cvec3d& maxp() const { return _max; };

  // leaf
  bool isLeaf() const { return _isLeaf; };

  // edge size
  int size() const { return _edges.size(); };
  vector<DualEdge*>& edges() { return _edges; };

  // insert an edge
  void insert(DualEdge *e) { _edges.push_back(e); };

  // split into 8
  void split();

  // find silhouettes
  void findSilhouettes(const cvec3d& n, double d, Silhouettes *sil);

  // boundary edges
  static vector<BoundaryEdge> boundary;

  // build the tree
  static OcTree* buildTree(TMesh *mesh);

private:
  // depth in octree
  int _depth;

  // bounding box
  cvec3d _min, _max;

  // leaf ?
  bool _isLeaf;

  // children
  OcTree *_child[8];

  // edges
  vector<DualEdge*> _edges;

  // the camera plane
  cvec3d _normal;
  double _d;

  // touch the camera plane
  bool touchPlane(const cvec3d& n, double d);

  // static members
  static const int MAX_EDGE_NUMBER;
};

#endif /* __DUALMESH_H__ */
